# -*- coding: UTF-8 -*-
import Queue
import atexit
import fnmatch
import json
import multiprocessing
import os
import signal
import startReader
import tempfile
import threading
import time
import traceback

from common import logger, config
from artix.common import servicelock
from unload.service.reader import Reader
from prettytable import PrettyTable, ALL
from httplib import INTERNAL_SERVER_ERROR
from bottle import get, post, response, request, Bottle, run

log = logger.getUnloadLogger('main')


class UnloadTypeEnum(set):
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError


UnloadTypes = UnloadTypeEnum(["TASK", "INTERVAL", "FLAG"])


def task_unload(shop, work_directory, cash, task=None, file_flag=None, converterName=None, closedShifts=0, type=None):
    # Выгрузка продаж по файлу-флагу/интервалу/задаче из веба КЦ. Метод не в классе, т.к. иначе его нельзя использовать
    # в мультипроцессинге
    log_str_thread_name = "Поток [%s] - " % multiprocessing.current_process().name if multiprocessing.current_process().name != 'MainProcess' else ''
    log.debug("%sСтарт выгрузки продаж для узла: магазин=%s, касса=%s" % (log_str_thread_name, shop.id, cash.cashId))
    result = startReader.unload(shop, work_directory, cashCode=cash.code,
                                shifts=task.get('shifts') if task else None,
                                dates=task.get('dates') if task else None,
                                granularity=task.get('granularity', 'document') if task else None,
                                closedShifts=task.get('closedShifts', 0) if task else closedShifts,
                                changedShifts=task.get('changedShifts') if task else False,
                                converterName=converterName,
                                request=file_flag)
    log.debug("%sЗавершена выгрузка для узла:  магазин=%s, касса=%s, результат=%s" % (
        log_str_thread_name, shop.id, cash.cashId, result))
    return result, cash.cashId if cash else None, type


class UnloadService(Bottle):

    def __init__(self):
        super(UnloadService, self).__init__()
        get("/exchangers/sales/unload/queue")(self.__get_unload_tasks_queue_status)
        post("/exchangers/sales/unload")(self.__add_unload_task_to_queue)
        self.__readConfig()
        self.nodes = []
        lockFile = os.path.normpath('%s/locks/lockArtixUnloadService' % tempfile.gettempdir())
        self.srvLock = servicelock.ServiceLock(lockFile, log)
        self.reader = Reader()
        self.unloadTasks = Queue.PriorityQueue(5000)
        atexit.register(self.srvLock.unlock)
        signal.signal(signal.SIGTERM, self.srvLock.unlock)
        self.timeoutNodeReread = 300  # таймаут перечитывания настроек
        self.notProcessedShops = []
        self.unloadTask = None
        self.taskCountPass = 0
        self.not_processed_cashes = []
        self.all_cashes_task = 0
        self.all_cashes_interval = 0
        self.all_cashes_flag = 0
        self.processed_cashes_task = 0
        self.processed_cashes_interval = 0
        self.processed_cashes_flag = 0
        self.current_cash = None
        self.unloadTaskPriority = 10

    def __start(self):
        log.info('Запуск REST-сервиса выгрузки продаж в АСТУ, host = %s, port = %s' % (
            self.restServiceHost, self.restServicePort))
        try:
            if not self.srvLock.lock(False):
               log.error("Сервис уже запущен")
               os._exit(1)  # выход без выполнения atexit
            t = threading.Thread(target=run,
                                 kwargs={'host': self.restServiceHost, 'port': self.restServicePort, 'debug': True})
            t.daemon = True
            t.start()
        except Exception as e:
            log.error(u'Ошибка: %s' % e)
            log.error(traceback.format_exc())

    def __readConfig(self):
        self.unloadCurrentSalesInterval = config.getInt('UNLOAD', 'unloadCurrentSalesInterval', 600)
        self.unloadSalesInterval = config.getInt('UNLOAD', 'unloadSalesInterval', 20)
        self.closedShiftsByFileFlag = config.getInt('UNLOAD', 'closedShiftsByFileFlag', 2)
        self.workDirectory = config.getString('main', 'workDirectory',
                                              '%s/artixExchange/workDir' % tempfile.gettempdir())
        self.multi_processing = config.getBool('UNLOAD', 'multiprocessing', False)
        self.processes_in_pool_count = config.getInt('UNLOAD', 'processCount', 10)
        self.restServiceHost = config.getString('UNLOAD', 'restHost', '0.0.0.0')
        self.restServicePort = config.getString('UNLOAD', 'restPort', 28082)

        log.debug('main.workDirectory - %s' % self.workDirectory)
        log.debug('UNLOAD.unloadCurrentSalesInterval - %s' % self.unloadCurrentSalesInterval)
        log.debug('UNLOAD.unloadSalesInterval - %s' % self.unloadSalesInterval)
        log.debug('UNLOAD.closedShiftsByFileFlag - %s' % self.closedShiftsByFileFlag)
        log.debug('UNLOAD.multiprocessing - %s' % self.multi_processing)
        log.debug('UNLOAD.processesInPool - %s' % self.processes_in_pool_count)

    def on_start(self):
        unloadCurrentSalesInterval = self.unloadCurrentSalesInterval
        unloadSalesInterval = self.unloadSalesInterval
        timeNodeReread = 0
        timeUnloadCurrentSales = 0
        timeFindRequest = 0
        nodeGetter = None

        try:
            nodeGetter = __import__('unload.node.%s.nodeGetter' % config.getString('UNLOAD', 'nodeGetter', 'cs30'),
                                    fromlist='q')
            self.nodes = nodeGetter.getNodes()
            timeNodeReread = time.time()
        except Exception as e:
            log.error(u'Ошибка получения нод. %s' % str(e))
            log.error(traceback.format_exc())
        self.__start()
        passCountSending = 0
        while True:
            if (time.time() - self.timeoutNodeReread) > timeNodeReread:
                timeNodeReread = time.time()
                try:
                    newNodes = nodeGetter.getNodes()
                    self.nodes = newNodes
                except:
                    pass
            #### Выгрузка текущих продаж по интервалу
            if unloadCurrentSalesInterval != 0:
                if (time.time() - unloadCurrentSalesInterval) > timeUnloadCurrentSales:
                    timeUnloadCurrentSales = time.time()
                    self.__unload_by_interval()
            ##################################################################
            #### Выгрузка по таску из REST'а
            self.__unload_by_task(self.nodes)
            ##################################################################
            #### Выгрузка по файлу-флагу
            if timeFindRequest == 0 or (time.time() - unloadSalesInterval) > timeFindRequest:
                timeFindRequest = time.time()
                self.__unload_by_flag()
            ##################################################################
            if passCountSending < 5:
                passCountSending += 1
            else:
                passCountSending = 0
                for node in self.nodes:
                    if self.__findFile(self.workDirectory + '/%s/unload' % node.id, 'ready'):
                        self.reader.send(node, self.workDirectory)
                    if self.__findFile(self.workDirectory + '/%s/unload' % node.id, 'readyOnline'):
                        self.reader.send(node, self.workDirectory, closedShifts=1)
            time.sleep(2)

    @staticmethod
    def __findFile(directory, pattern):
        if os.path.exists(directory):
            for file in os.listdir(directory):
                if fnmatch.fnmatch(file.lower(), pattern.lower()):
                    return os.path.join(directory, file)
        return None

    def __unload_by_interval(self):
        cashes_to_unload = []
        for node in self.nodes:
            for cash in node.cashes:
                for converter in cash.converters:
                    if hasattr(converter, 'onlineSalePath') and converter.onlineSalePath:
                        # собираем данные для выгрузки продаж - список словарей с параметрами для передачи в метод task_unload
                        cashes_to_unload.append({'shop': node, 'work_directory': self.workDirectory, 'cash': cash,
                                                 'converterName': converter.name, 'closedShifts': 1,
                                                 'type': UnloadTypes.INTERVAL})
        # запускается выгрузка, если нашли кассы
        if len(cashes_to_unload) > 0:
            log.info(u'Выгрузка текущих продаж...')
            self.all_cashes_interval = len(cashes_to_unload)
            self.processed_cashes_interval = 0
            self.__unload(cashes_to_unload)

    def __unload_by_flag(self):
        cashes_to_unload = []
        for node in self.nodes:
            for cash in node.cashes:
                for converter in cash.converters:
                    if converter.flag:
                        requestFile = self.__findFile(os.path.dirname(converter.flag), os.path.basename(converter.flag))
                        if requestFile:
                            # собираем данные для выгрузки продаж - список словарей с параметрами для передачи в метод task_unload
                            # предполагается, что флаг будет уникальным для каждой кассы
                            cashes_to_unload.append({'shop': node, 'work_directory': self.workDirectory, 'cash': cash,
                                                     'converterName': converter.name, 'file_flag': requestFile,
                                                     'closedShifts': self.closedShiftsByFileFlag,
                                                     'type': UnloadTypes.FLAG})
        # запускается выгрузка, если нашли кассы
        if len(cashes_to_unload) > 0:
            log.info(u'Выгрузка продаж по запросу файла-флага')
            self.all_cashes_flag = len(cashes_to_unload)
            self.processed_cashes_flag = 0
            self.__unload(cashes_to_unload)

    def __unload_by_task(self, nodes):
        # пытаемся выполнить задачу на выгрузку не больше 20 раз, затем пропускаем ее
        self.processed_cashes_task = 0
        self.current_cash = None
        if self.taskCountPass > 20:
            log.warning('Невозможно выполнить выгрузку продаж для касс %s' % self.not_processed_cashes)
            log.warning('Кол-во попыток выгрузки истекло. Приступаем к следующей задаче...')
            del self.not_processed_cashes[:]
            self.__delete_task_from_queue(self.unloadTaskPriority, self.unloadTask)

        if len(self.not_processed_cashes) == 0:
            self.taskCountPass = 0
            try:
                if self.unloadTasks.unfinished_tasks > 0:
                    self.unloadTaskPriority, self.unloadTask = self.unloadTasks.queue.__getitem__(0)
                    log.info('Сервис начал выполнять задачу %s' % self.unloadTask)
                    # собираем магазины из задачи или берем все
                    self.notProcessedShops = [str(self.unloadTask.get('shopId'))] if (self.unloadTask.get('shopId')) else [str(id) for id in self.unloadTask.get('shopIds')] if (self.unloadTask.get('shopIds')) else [str(n.id) for n in nodes]
                    log.info('Найдены магазины: %s' % self.notProcessedShops)
                    # собираем кассы для выгрузки по отобранным магазинам
                    for shop in filter(lambda x: x.id in self.notProcessedShops, nodes):
                        for cash in shop.cashes:
                            # случай выгрузки в групповых операциях по кассам
                            if self.unloadTask.get('cashIds'):
                                if cash.cashId in self.unloadTask.get('cashIds'):
                                    self.not_processed_cashes.append(cash.cashId)
                                else:
                                    continue
                            # случай, когда выгрузка запущена на одной кассе
                            elif self.unloadTask.get('cashCode'):
                                if cash.code == self.unloadTask.get('cashCode'):
                                    self.not_processed_cashes.append(cash.cashId)
                            else:
                                # случай выгрузки по магазинам, когда прогружаем все кассы
                                self.not_processed_cashes.append(cash.cashId)
                    if len(self.not_processed_cashes) == 0:
                        log.info('Не найдено касс для выгрузки продаж, задача будет удалена из очереди')
                        self.__delete_task_from_queue(self.unloadTaskPriority, self.unloadTask)
                    else:
                        log.info('Найдены кассы для выгрузки продаж: %s' % self.not_processed_cashes)
                else:
                    self.unloadTask = None
            except:
                self.unloadTask = None
        else:
            self.taskCountPass += 1
            log.info("Не удалось выполнить выгрузку для касс: %s. Пробуем снова" % self.not_processed_cashes)
        if self.unloadTask and len(self.not_processed_cashes) > 0:
            # собираем идентификаторы касс для выгрузки продаж
            self.all_cashes_task = len(self.not_processed_cashes)
            # собираем данные для выгрузки продаж - список словарей с параметрами для передачи в метод task_unload
            cashes_to_unload = []
            for shop in filter(lambda x: x.id in self.notProcessedShops, nodes):
                for cash in filter(lambda x: x.cashId in self.not_processed_cashes, shop.cashes):
                    cashes_to_unload.append(
                        {'shop': shop, 'work_directory': self.workDirectory, 'task': self.unloadTask, 'cash': cash,
                         'type': UnloadTypes.TASK})
            self.__unload(cashes_to_unload)

    def __unload(self, data):
        if self.multi_processing:
            log.debug('Запущена выгрузка в многопоточном режиме')
            # запускаем выгрузку для каждой кассы в отдельном процессе
            pool = multiprocessing.Pool(self.processes_in_pool_count)
            # эта штука работает как-то так:
            # 1. создается набор процессов в кол-ве self.processes_in_pool_count
            # 2. каждый процесс берет задачу и выполняет
            # 3. когда задача выполнена, процесс отдает результат в callback(там делаем нашу обработку) и берет другую задачу, если они еще есть
            # 4. когда все задачи выполнены, процессы завершаются
            for kwargs in data:
                pool.apply_async(func=task_unload, kwds=kwargs, callback=self.__post_unload)
            pool.close()
            pool.join()
        else:
            for kwargs in data:
                result = task_unload(**kwargs)
                self.__post_unload(result)

    def __post_unload(self, unload_result):
        """
        Обработка результата выгрузки продаж, подсчет успешно/неуспешно прогруженных касс
        """
        res, cash_id, type = unload_result[0], unload_result[1], unload_result[2]
        self.current_cash = cash_id
        log.debug('Текущая касса: %s' % self.current_cash)
        # TODO подумать, как получше организовать структуру разных выгрузок
        #  выглядит, как абстрактный класс с методами singe_unload(), multi_unload(), post_process()
        #  и от него 3 наследника с разными типами выгрузки и со своими обработками и т.д.
        if type == UnloadTypes.TASK:
            self.processed_cashes_task += 1
            log.info("Обработано касс: %s из %s" % (self.processed_cashes_task, self.all_cashes_task))
            if res and self.unloadTask:
                try:
                    self.not_processed_cashes.remove(cash_id)
                except Exception as e:
                    log.error(u'Ошибка удаления кассы %s из списка для выгрузки: %s' % (cash_id, e))
                # после того, как все кассы прогрузились, удаляем задачу из очереди
            if len(self.not_processed_cashes) == 0 and self.unloadTask:
                log.info("Сервис закончил выполнять задачу")
                self.__delete_task_from_queue(self.unloadTaskPriority, self.unloadTask)
        elif type == UnloadTypes.INTERVAL:
            self.processed_cashes_interval += 1
            log.info("Обработано касс: %s из %s" % (self.processed_cashes_interval, self.all_cashes_interval))
            if self.processed_cashes_interval == self.all_cashes_interval:
                log.info("Сервис закончил выполнять выгрузку текущих продаж")
        elif type == UnloadTypes.FLAG:
            self.processed_cashes_flag += 1
            log.info("Обработано касс: %s из %s" % (self.processed_cashes_flag, self.all_cashes_flag))
            if self.processed_cashes_flag == self.all_cashes_flag:
                log.info("Сервис закончил выполнять выгрузку продаж по запросу файла-флага")

    def __delete_task_from_queue(self, priority, task):
        """
        Удаление задачи из очереди
        """
        try:
            self.unloadTasks.task_done()  # нужно, чтобы у очереди изменился признак unfinished_tasks
            self.unloadTasks.queue.remove((priority, task))
        except Exception as e:
            log.error(u'При удалении задачи из очереди произошла ошибка: %s' % e)

    def __add_unload_task_to_queue(self):
        """
        Добавляет задачу в очередь
        """
        response.content_type = 'application/json'
        task = None
        try:
            task = json.loads(request.body.read(), encoding='utf-8')
            priority = task.pop('priority', 10)
            self.unloadTasks.put((priority, task), block=True, timeout=60)
            mess = u'Task %s added to queue' % task
            log.info(mess)
            result = {u'resultCode': 0, u'info': mess}
        except Queue.Full:
            mess = u'Queue is full. Can not add task %s to queue.' % task
            log.warning(mess)
            result = {u'resultCode': 10, u'info': mess}
        except Exception, e:
            log.warning(u'Ошибка REST сервиса. %s' % str(e))
            log.warning(traceback.format_exc())
            response.status = INTERNAL_SERVER_ERROR
            result = {u'resultCode': 100, u'info': str(e)}
        return json.dumps(result, ensure_ascii=False)

    def __get_unload_tasks_queue_status(self):
        """
        Возвращает таблицу с данными по задачам из очереди
        """
        try:
            queue_table = PrettyTable()
            queue_table.field_names = ['Индекс в очереди', 'Приоритет', 'Магазины/Кассы', 'Смены', 'Даты',
                                       'Тип выгрузки', 'Закрытые смены', 'Измененные смены', 'Прогресс', 'Текущий узел']
            # так называемые "стили" таблицы: выравнивание полей по левому краю
            # плюс ограничение длины для полей с магазинами и кассами
            queue_table.align = "l"
            queue_table._max_width = {queue_table.field_names[2]: 90}
            queue_table.hrules = ALL

            if len(self.unloadTasks.queue) > 0:
                for i in range(len(self.unloadTasks.queue)):
                    priority, task = self.unloadTasks.queue.__getitem__(i)
                    shops = task.get('shopId') if task.get('shopId') else ','.join(task.get('shopIds')) if (task.get('shopIds')) else None
                    cashes = task.get('cashCode') if task.get('cashCode') else ','.join(task.get('cashIds')) if (task.get('cashIds')) else None
                    progress = str("%d/%d" % (self.processed_cashes_task, self.all_cashes_task))
                    row = [i, priority, shops if shops else cashes, task.get('shifts'), task.get('dates'), task.get('granularity'),
                           task.get('closedShifts'), task.get('changedShifts'), progress, '%s' % self.current_cash]
                    queue_table.add_row(row)
            return queue_table.get_string()
        except Exception as e:
            log.warning('Ошибка получения состояния очереди: %s' % str(e))
            result = {u'resultCode': 500, u'info': str(e)}
            return json.dumps(result, ensure_ascii=False)
